from unittest.mock import AsyncMock, MagicMock

import pytest
from jinja2 import Environment
from jinja2.bccache import Bucket
from jinja2_async_environment.bccache import AsyncBytecodeCache


class TestAsyncBytecodeCache:
    def test_get_cache_key(self) -> None:
        cache = MagicMock(spec=AsyncBytecodeCache)
        cache.get_cache_key.side_effect = lambda name, filename: filename or name
        assert cache.get_cache_key("name", "filename") == "filename"
        assert cache.get_cache_key("name", None) == "name"

    def test_get_source_checksum(self) -> None:
        cache = MagicMock(spec=AsyncBytecodeCache)
        cache.get_source_checksum.side_effect = lambda source: str(hash(source))
        source = "template source"
        checksum = cache.get_source_checksum(source)
        assert isinstance(checksum, str)
        assert checksum == str(hash(source))
        different_source = "different template source"
        different_checksum = cache.get_source_checksum(different_source)
        assert checksum != different_checksum

    @pytest.mark.asyncio
    async def test_not_implemented_methods(self) -> None:
        cache = MagicMock(spec=AsyncBytecodeCache)
        bucket = MagicMock(spec=Bucket)
        environment = MagicMock(spec=Environment)
        cache.get_bucket_name.side_effect = NotImplementedError
        cache.load_bytecode = AsyncMock(side_effect=NotImplementedError)
        cache.dump_bytecode = AsyncMock(side_effect=NotImplementedError)
        cache.get_bucket = AsyncMock(side_effect=NotImplementedError)
        cache.set_bucket = AsyncMock(side_effect=NotImplementedError)
        with pytest.raises(NotImplementedError):
            cache.get_bucket_name("key")
        with pytest.raises(NotImplementedError):
            await cache.load_bytecode(bucket)
        with pytest.raises(NotImplementedError):
            await cache.dump_bytecode(bucket)
        with pytest.raises(NotImplementedError):
            await cache.get_bucket(environment, "name", "filename", "source")
        with pytest.raises(NotImplementedError):
            await cache.set_bucket(bucket)



import typing as t

from jinja2 import nodes
from jinja2.compiler import CodeGenerator, CompilerExit, Frame


class AsyncFrame(Frame):
    block_frame: "AsyncFrame | None"
    block_counters: t.Mapping[str, int]
    block_frame_id: int
    require_output_check: bool
    has_known_extends: bool
    toplevel: bool
    rootlevel: bool
    buffer: t.Any | None

    def copy(self) -> t.Self:
        rv = super().copy()
        t.cast(AsyncFrame, rv)
        return rv

    def inspect(self, nodes: t.Any | None = None) -> None: ...

    def push_scope(self) -> None: ...

    def pop_scope(self) -> None: ...

    def find_break(self) -> bool: ...

    def find_continue(self) -> bool: ...


class AsyncCodeGenerator(CodeGenerator):
    environment: t.Any
    name: str
    filename: str
    stream: t.Any
    extends_so_far: int
    has_known_extends: bool

    def visit_Block(self, node: nodes.Block, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        level = 0
        if frame.toplevel:
            if self.has_known_extends:
                return
            if self.extends_so_far > 0:
                self.writeline("if parent_template is None:")
                self.indent()
                level += 1
        if node.scoped:
            context = self.derive_context(frame)
        else:
            context = self.get_context_ref()
        if node.required:
            self.writeline(f"if len(context.blocks[{node.name!r}]) <= 1:", node)
            self.indent()
            self.writeline(
                f'raise TemplateRuntimeError("Required block {node.name!r} not found")',
                node,
            )
            self.outdent()
        if not self.environment.enable_async and frame.buffer is None:
            self.writeline(
                f"yield from context.blocks[{node.name!r}][0]({context})", node
            )
        else:
            self.writeline(
                f"{self.choose_async()}for event in context.blocks[{node.name!r}][0]({context}):",
                node,
            )
            self.indent()
            self.simple_write("event", frame)
            self.outdent()
        self.outdent(level)

    def visit_Extends(self, node: nodes.Extends, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        if not frame.require_output_check:
            raise CompilerExit()
        if not frame.toplevel:
            self.fail("cannot use extend from a non top-level scope", node.lineno)
        if self.extends_so_far > 0:
            if not self.has_known_extends:
                self.writeline("if parent_template is not None:")
                self.indent()
            self.writeline('raise TemplateRuntimeError("extended multiple times")')
            if self.has_known_extends:
                raise CompilerExit()
            else:
                self.outdent()
        self.writeline("parent_template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.writeline("for name, parent_block in parent_template.blocks.items():")
        self.indent()
        self.writeline("context.blocks.setdefault(name, []).append(parent_block)")
        self.outdent()
        if frame.rootlevel:
            self.has_known_extends = True
        self.extends_so_far += 1

    def visit_Include(self, node: nodes.Include, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        if node.ignore_missing:
            self.writeline("try:")
            self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        if node.ignore_missing:
            self.outdent()
            self.writeline("except TemplateNotFound:")
            self.indent()
            self.writeline("pass")
            self.outdent()
            self.writeline("else:")
            self.indent()
        if node.with_context:
            self.writeline(
                f"async for event in template.root_render_func(template.new_context(context.get_all(), True, {self.dump_local_context(frame)})):"
            )
        else:
            self.writeline(
                "async for event in (await template._get_default_module_async())._body_stream:"
            )
        self.indent()
        self.simple_write("event", frame)
        self.outdent()
        if node.ignore_missing:
            self.outdent()

    def _import_common(
        self, node: nodes.Import | nodes.FromImport, frame: Frame
    ) -> None:
        frame = t.cast(AsyncFrame, frame)
        self.writeline("try:")
        self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.outdent()
        self.writeline("except TemplateNotFound:")
        self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.outdent()



import importlib.util
import typing as t
import zipimport
from contextlib import suppress
from importlib import import_module
from pathlib import Path

from aiopath import AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2.loaders import BaseLoader
from jinja2.utils import internalcode
from .environment import AsyncEnvironment


class PackageSpecNotFound(TemplateNotFound): ...


class LoaderNotFound(TemplateNotFound): ...


SourceType = tuple[
    str | bytes, str | None, t.Callable[[], bool | t.Awaitable[bool]] | None
]


class AsyncLoaderProtocol(t.Protocol):
    async def async_get_source(self, template: str | AsyncPath) -> SourceType: ...

    async def async_list_templates(self) -> list[str]: ...

    async def async_load(
        self,
        environment: AsyncEnvironment,
        name: str,
        env_globals: dict[str, t.Any] | None = None,
    ) -> Template: ...


class AsyncBaseLoader(BaseLoader):
    has_source_access: bool = True
    searchpath: list[AsyncPath]

    def __init__(self, searchpath: AsyncPath | t.Sequence[AsyncPath]) -> None:
        if isinstance(searchpath, AsyncPath):
            self.searchpath = [searchpath]
        elif isinstance(searchpath, (list, tuple)):
            self.searchpath = list(searchpath)
        else:
            raise TypeError(
                "searchpath must be an AsyncPath or a sequence of AsyncPath objects"
            )

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )
        raise TemplateNotFound(template_path.name)

    async def async_list_templates(self) -> list[str]:
        raise TypeError("this loader cannot iterate over all templates")

    @internalcode
    async def async_load(
        self,
        environment: AsyncEnvironment,
        name: str,
        env_globals: dict[str, t.Any] | None = None,
    ) -> Template:
        if env_globals is None:
            env_globals = {}
        source, path, uptodate = await self.async_get_source(name)
        source_str = source.decode("utf-8") if isinstance(source, bytes) else source
        bcc = environment.bytecode_cache
        bucket = None
        if bcc:
            bucket = await bcc.async_get_bucket(environment, name, path, source_str)
            code = bucket.code
        else:
            code = None
        if not code:
            if path is None:
                code = environment.compile(source_str, name)
            else:
                code = environment.compile(source_str, name, path)
        if bcc and bucket is not None and (not bucket.code):
            bucket.code = code
            await bcc.async_set_bucket(bucket)
        return environment.template_class.from_code(
            environment,
            code,
            env_globals,
            t.cast(t.Optional[t.Callable[[], bool]], uptodate),
        )


class AsyncFileSystemLoader(AsyncBaseLoader):
    encoding: str
    followlinks: bool

    def __init__(
        self,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        encoding: str = "utf-8",
        followlinks: bool = False,
    ) -> None:
        super().__init__(searchpath)
        self.encoding = encoding
        self.followlinks = followlinks

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )
        path: AsyncPath | None = None
        for sp in self.searchpath:
            candidate = sp / template_path
            if await candidate.is_file():
                path = candidate
                break
        if path is None:
            raise TemplateNotFound(template_path.name)
        try:
            resp = await path.read_bytes()
        except FileNotFoundError:
            raise TemplateNotFound(path.name)
        mtime = (await path.stat()).st_mtime

        async def _uptodate() -> bool:
            try:
                return (await path.stat()).st_mtime == mtime
            except OSError:
                return False

        return (
            resp.decode(self.encoding) if isinstance(resp, bytes) else resp,
            str(path),
            _uptodate,
        )

    async def async_list_templates(self) -> list[str]:
        results: set[str] = set()
        for sp in self.searchpath:
            results.update({str(p) async for p in sp.rglob("*.html")})
        return sorted(results)


class AsyncPackageLoader(AsyncBaseLoader):
    package_path: AsyncPath
    package_name: str
    encoding: str
    _loader: t.Any
    _archive: str | None
    _template_root: AsyncPath

    def __init__(
        self,
        package_name: str,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        package_path: AsyncPath = AsyncPath("templates"),
        encoding: str = "utf-8",
    ) -> None:
        super().__init__(searchpath)
        self.package_path = package_path
        self.package_name = package_name
        self.encoding = encoding
        import_module(package_name)
        spec = importlib.util.find_spec(package_name)
        if not spec:
            raise PackageSpecNotFound("An import spec was not found for the package")
        loader = spec.loader
        if not loader:
            raise LoaderNotFound("A loader was not found for the package")
        self._loader = loader
        self._archive = None
        template_root: AsyncPath | None = None
        if isinstance(loader, zipimport.zipimporter):
            self._archive = loader.archive
            pkg_locations = spec.submodule_search_locations or []
            if pkg_locations:
                pkgdir = next(iter(pkg_locations))
                template_root = AsyncPath(pkgdir) / package_path
        else:
            roots: list[Path] = []
            if spec.submodule_search_locations:
                roots.extend([Path(s) for s in spec.submodule_search_locations])
            elif spec.origin is not None:
                roots.append(Path(spec.origin))
            for root in roots:
                candidate = root / package_path
                if candidate.is_dir():
                    template_root = AsyncPath(root)
                    break
        if not template_root:
            raise ValueError(
                f"The {package_name!r} package was not installed in a way that PackageLoader understands"
            )
        self._template_root = template_root

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )
        path = self._template_root / template_path
        if self._archive:
            if not await path.is_file():
                raise TemplateNotFound(path.name)
            source_bytes = await path.read_bytes()
            mtime = (await path.stat()).st_mtime

            async def _uptodate() -> bool:
                return await path.is_file() and (await path.stat()).st_mtime == mtime

            return (source_bytes.decode(self.encoding), str(path), _uptodate)
        else:
            try:
                source_bytes = self._loader.get_data(str(path))
            except OSError as exc:
                raise TemplateNotFound(path.name) from exc
            uptodate: t.Any = None
            return (source_bytes.decode(self.encoding), str(path), uptodate)

    async def async_list_templates(self) -> list[str]:
        results: list[str] = []
        if self._archive is None:
            paths = self._template_root.rglob("*.html")
            results.extend([str(p) async for p in paths])
        else:
            if not hasattr(self._loader, "_files"):
                raise TypeError(
                    "This zip import does not have the required metadata to list templates"
                )
            prefix = self._template_root.name
            for name in self._loader._files.keys():
                if name.startswith(prefix) and await AsyncPath(name).is_file():
                    results.append(name)
        results.sort()
        return results


class AsyncDictLoader(AsyncBaseLoader):
    mapping: t.Mapping[str, str]

    def __init__(
        self,
        mapping: t.Mapping[str, str],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.mapping = mapping

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        template_name: str = (
            template.name if isinstance(template, AsyncPath) else template
        )
        if template_name in self.mapping:
            source = self.mapping[template_name]
            return (source, None, lambda: source == self.mapping.get(template_name))
        raise TemplateNotFound(template_name)

    async def async_list_templates(self) -> list[str]:
        return sorted(list(self.mapping))


class AsyncFunctionLoader(AsyncBaseLoader):
    load_func: t.Callable[
        [str | AsyncPath], t.Awaitable[SourceType] | SourceType | None
    ]

    def __init__(
        self,
        load_func: t.Callable[
            [str | AsyncPath], t.Awaitable[SourceType] | SourceType | None
        ],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.load_func = load_func

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        result = self.load_func(template)
        if result is None:
            template_name: str = (
                template.name if isinstance(template, AsyncPath) else template
            )
            raise TemplateNotFound(template_name)
        if isinstance(result, tuple):
            return result
        if hasattr(result, "__await__"):
            awaited_result = await result
            return awaited_result
        if isinstance(result, str):
            template_str = str(template)
            return (result, template_str, lambda: True)
        raise TypeError(f"Unexpected source type: {type(result)}")


class AsyncChoiceLoader(AsyncBaseLoader):
    loaders: list[AsyncBaseLoader]

    def __init__(
        self,
        loaders: list[AsyncBaseLoader],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.loaders = loaders

    async def async_get_source(self, template: str | AsyncPath) -> SourceType:
        for loader in self.loaders:
            with suppress(TemplateNotFound):
                return await loader.async_get_source(template)
        template_name: str = (
            template.name if isinstance(template, AsyncPath) else template
        )
        raise TemplateNotFound(template_name)

    async def async_list_templates(self) -> list[str]:
        found: set[str] = set()
        for loader in self.loaders:
            found.update(await loader.async_list_templates())
        return sorted(found)



import typing as t
from contextlib import suppress
from weakref import ref

from jinja2 import Environment
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound, TemplatesNotFound, UndefinedError
from jinja2.runtime import Undefined
from jinja2.utils import internalcode
from .bccache import AsyncBytecodeCache
from .compiler import AsyncCodeGenerator, CodeGenerator


class AsyncEnvironment(Environment):
    code_generator_class: t.Type[CodeGenerator] = AsyncCodeGenerator
    loader: t.Any | None = None
    bytecode_cache: AsyncBytecodeCache | None = None
    enable_async = True

    @internalcode
    def get_template(
        self,
        name: str | Template,
        parent: str | Template | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use get_template_async instead")

    @internalcode
    async def get_template_async(
        self,
        name: str | Template,
        parent: str | Template | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(name, Template):
            return name
        if parent is not None:
            name = self.join_path(str(name), str(parent))
        return await self._load_template_async(name, globals)

    @internalcode
    def select_template(
        self,
        names: t.Iterable[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use select_template_async instead")

    @internalcode
    async def select_template_async(
        self,
        names: t.Iterable[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(names, Undefined):
            names._fail_with_undefined_error()
        if not names:
            raise TemplatesNotFound(
                message="Tried to select from an empty list of templates."
            )
        names_list = []
        for name in names:
            if isinstance(name, Template):
                return name
            if parent is not None:
                name = self.join_path(str(name), parent)
            with suppress(TemplateNotFound, UndefinedError):
                return await self._load_template_async(name, globals)
            names_list.append(str(name))
        raise TemplatesNotFound(names_list)

    @internalcode
    def get_or_select_template(
        self,
        template_name_or_list: str | Template | t.Sequence[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use get_or_select_template_async instead")

    @internalcode
    async def get_or_select_template_async(
        self,
        template_name_or_list: str | Template | t.Sequence[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(template_name_or_list, (str, Undefined)):
            return await self.get_template_async(template_name_or_list, parent, globals)
        elif isinstance(template_name_or_list, Template):
            return template_name_or_list
        return await self.select_template_async(template_name_or_list, parent, globals)

    @internalcode
    async def _load_template_async(
        self,
        name: str | Template | t.Iterable[str | Template],
        globals: t.MutableMapping[str, t.Any] | None,
    ) -> Template:
        if isinstance(name, Template):
            return name
        if isinstance(name, str):
            return await self._get_template(name, globals)
        names_list = []
        for template_name in name:
            if isinstance(template_name, Template):
                return template_name
            with suppress(TemplateNotFound):
                return await self._get_template(str(template_name), globals)
            names_list.append(str(template_name))
        raise TemplatesNotFound(names_list)

    async def _get_template(
        self, name: str, globals: t.MutableMapping[str, t.Any] | None
    ) -> Template:
        if self.loader is None:
            raise TypeError("no loader for this environment specified")
        cache_key = (ref(self.loader), name)
        if self.cache is not None:
            with suppress(TypeError):
                template = self.cache.get(cache_key)
                if template is not None and (
                    not self.auto_reload or template.is_up_to_date
                ):
                    if globals:
                        template.globals.update(globals)
                    return template
        template = await self.loader.async_load(self, name, self.make_globals(globals))
        if self.cache is not None:
            self.cache[cache_key] = template
        return template



import typing as t
from types import MappingProxyType

from jinja2 import Environment
from jinja2.bccache import Bucket, BytecodeCache
from redis.asyncio import Redis, RedisCluster  # type: ignore


class AsyncBytecodeCache(BytecodeCache):
    def load_bytecode(self, bucket: Bucket) -> None:
        raise NotImplementedError(
            "AsyncBytecodeCache.load_bytecode must be implemented."
        )

    def dump_bytecode(self, bucket: Bucket) -> None:
        raise NotImplementedError(
            "AsyncBytecodeCache.dump_bytecode must be implemented."
        )

    def get_bucket_name(self, key: str) -> str:
        raise NotImplementedError("get_bucket_name must be implemented.")

    async def async_get_bucket(
        self, environment: Environment, name: str, filename: str | None, source: str
    ) -> Bucket:
        raise NotImplementedError(
            "AsyncBytecodeCache.async_get_bucket must be implemented."
        )

    async def async_set_bucket(self, bucket: Bucket) -> None:
        raise NotImplementedError(
            "AsyncBytecodeCache.async_set_bucket must be implemented."
        )


class AsyncRedisBytecodeCache(AsyncBytecodeCache):
    prefix: str | None
    client: Redis | RedisCluster  # type: ignore
    configs: MappingProxyType[str, t.Any]

    def __init__(
        self,
        prefix: str | None = None,
        client: Redis | RedisCluster | None = None,  # type: ignore
        **configs: t.Any,
    ) -> None:
        self.prefix = prefix
        self.client = client or Redis(**configs)
        self.configs = MappingProxyType(configs)

    def get_cache_key(self, name: str, filename: str | None = None) -> str:
        return filename or name

    def get_source_checksum(self, source: str) -> str:
        return str(hash(source))

    def get_bucket_name(self, key: str) -> str:
        return ":".join([self.prefix, key]) if self.prefix else key

    async def load_bytecode(self, bucket: Bucket) -> t.Optional[bytes]:  # type: ignore[override]
        code = await self.client.get(self.get_bucket_name(bucket.key))  # type: ignore
        if code:
            bucket.bytecode_from_string(code)
            return code
        return None

    async def dump_bytecode(self, bucket: Bucket) -> None:  # type: ignore[override]
        await self.client.set(  # type: ignore
            self.get_bucket_name(bucket.key), bucket.bytecode_to_string()
        )

    async def get_bucket(  # type: ignore[override]
        self, environment: Environment, name: str, filename: str | None, source: str
    ) -> Bucket:
        key = self.get_cache_key(name, filename or name)
        checksum = self.get_source_checksum(source)
        bucket = Bucket(environment, key, checksum)
        await self.load_bytecode(bucket)
        return bucket

    async def set_bucket(self, bucket: Bucket) -> None:  # type: ignore[override]
        await self.dump_bytecode(bucket)

    async def async_get_bucket(
        self, environment: Environment, name: str, filename: str | None, source: str
    ) -> Bucket:
        return await self.get_bucket(environment, name, filename, source)

    async def async_set_bucket(self, bucket: Bucket) -> None:
        return await self.set_bucket(bucket)
